//tetris.c - Wilson Tetris for MS-DOS (IBM-PC) - 02/29/2020
//adapted from Wilson Tetris for TRS-80 v1- 05/28/2019
//this version uses BIOS int calls for cross-computer compatibility
//including the SANYO-MBC 550 which is not IBM-PC compatible
//Using assembly routines for faster access to keyboard and video
//INT 1aH for timing
//INT 10H for video
//INT 16H for keyboard
 
#include <stdio.h>
#include <stdlib.h>
#include <i86.h>

#define pfsizeX 10 //playfield size (in cells)
#define pfsizeY 22 //note that top 4 lines are not drawn
#define pfX 8 //upper left corner of playfield location (pixel)
#define pfY 3
#define npX 60 //upper left corner of next piece indicator
#define npY 2
#define clspX 2 //spacing between cell elements (pixels)
#define clspY 1

void controls(void);

unsigned short playfield[pfsizeX][pfsizeY];  //the game area
unsigned short cscreen[pfsizeX][pfsizeY - 4]; //a copy of what gets drawn on the screen (game area)
unsigned short pieceC[4][4]; //piece container
signed short pieceCX;  //location of upper left corner of piece container
signed short pieceCY;
unsigned short pieceT; //type of piece
unsigned short pieceR; //rotation of piece
unsigned short nextpieceT;
unsigned short nextpieceR;

unsigned short dcontrol; //locks in control every frame
unsigned short rcontrol;

unsigned short lastdcontrol = 0; //for "debouncing" of inputs
unsigned short lastrcontrol = 0;
unsigned short drepeatframe = 0;
#define drepeatframes 3 //wait _ frames before repeatedly going in one direction

unsigned short dropframe = 0;  //counter for number of frames between block drops
unsigned short level = 60; //LEVEL decreases frame drop from maxlevel frames to 0 frames (levels 0 to MAXLEVEL)
#define maxlevel 10

unsigned short ngame = 1; //when set, starts new game
unsigned short newpc = 0; //when set, get a new piece

unsigned int lines = 0; //NUMBER OF LINES CLEARED
long score = 0; //TOTAL SCORE (using NES rules)

unsigned short fdrop;   //number of blocks that piece has been fast dropped
unsigned short lslvi = 0; //lines since level increase (when this gets to 10, increase the level)
short lvladd = 0;
union REGS regs;

unsigned int readKeyA(); //read the last key in the keyboard buffer using BIOS int 16h
#pragma aux readKeyA =          \
	"mov bx, 0000h"             \
	"getkey:"                   \
    "mov ah, 01h"               \
    "int 16h"                   \
    "jz nokey"                  \
	"mov ah, 00h"               \
    "int 16h"                   \
	"mov bx, ax"                \
	"jp getkey"                 \
    "nokey:"                    \
    value [bx]                  \
    modify [ax bx];

void clearScreen(){
  int y = 0;
  for (y = 0; y <= 25; y++){
    printf("\r\n");
  }
}
void setCursor(unsigned short y, unsigned short x){
  regs.h.ah = 0x02;
  regs.h.bh = 0x00;
  regs.h.dl = x;
  regs.h.dh = y;
  int86(0x10, &regs, &regs);
}
void writeCharacter(unsigned short t){
  regs.h.ah = 0x0E;
  regs.h.bh = 0x00;
  regs.h.al = t;
  int86(0x10, &regs, &regs);
}


void setblock(short x, short y){ //uses pixel coordinates
  regs.h.ah = 0x02;
  regs.h.bh = 0x00;
  regs.h.dl = x;
  regs.h.dh = y;
  int86(0x10, &regs, &regs);
  regs.h.ah = 0x0A;
  regs.h.bh = 0x00;
  regs.h.al = 219;
  regs.w.cx = 2;
  int86(0x10, &regs, &regs);
}
void clearblock(short x, short y){
  regs.h.ah = 0x02;
  regs.h.bh = 0x00;
  regs.h.dl = x;
  regs.h.dh = y;
  int86(0x10, &regs, &regs);
  regs.h.ah = 0x0A;
  regs.h.bh = 0x00;
  regs.h.al = 32;
  regs.w.cx = 2;
  int86(0x10, &regs, &regs);
}

void cdraw(short x, short y, short color){ //only draws on screen if it needs to (uses cell coordinates)
  if (cscreen[x][y] != color){
    cscreen[x][y] = color;
    if (color != 0){
      setblock( (x)*clspX + pfX, (y)*clspY + pfY );
    } else {
      clearblock( (x)*clspX + pfX, (y)*clspY + pfY);
    }
  }
}
void recopyscreen(){ //copies playfield to screen
  short i = 0;
  for (i = 0; i < pfsizeX; i++){ //manually redraw the entire screen
    short j = 0;
    for (j = 4; j < pfsizeY; j++){
      cdraw(i, j - 4, playfield[i][j]);
    }
  }
  setCursor( 3, 47 ); printf("%ld\r\n", score);
  setCursor( 5, 47 ); printf("%d\r\n", level);
  setCursor( 7, 47 ); printf("%d\r\n", lines);
}
void controls(){
  unsigned short k = readKeyA()&0x00FF;
  while(k){
    if (k == '@'){
      ngame = 1;
    }
    if (dcontrol == 0){
      if (k == 'j'){
        dcontrol = 4;
      }
      if (k == 'l'){
        dcontrol = 6;
      }
      if (k == 'i'){
        dcontrol = 8;
      }
      if (k == 'k'){
        dcontrol = 2;
      }
      if (dcontrol == 0){
        drepeatframe = 0;
      }
      if (dcontrol == lastdcontrol){      //short delay before fast motion
        if (drepeatframe <= drepeatframes){
          drepeatframe++;
          dcontrol = 3; //lockout if within lockout period
        }
      }
    }
    if (rcontrol == 0){
      if (k == 'a'){
        rcontrol = 1;
      }
      if (k == 's'){
        rcontrol = 2;
      }
      if (rcontrol == lastrcontrol){
        rcontrol = 3;
      }
    }
  k = readKeyA()&&0x00FF;
  }
}
void clearControls(){
  if (rcontrol != 3){
    lastrcontrol = rcontrol;
  }
  if (dcontrol != 3){
    lastdcontrol = dcontrol;
  }
  rcontrol = 0;
  dcontrol = 0;
}

void loadpiece(){ //hardcoded all pieces
  switch (pieceT){
    case 1: //long one
      switch (pieceR){
        case 1:
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 1; pieceC[1][2] = 1; pieceC[2][2] = 1; pieceC[3][2] = 1;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 2:
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 1; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 1; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 0; pieceC[2][2] = 1; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 1; pieceC[3][3] = 0;
          break;
      }
      break;
    case 2: //backwards L
      switch (pieceR){
        case 1:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 2; pieceC[1][2] = 2; pieceC[2][2] = 2; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 2; pieceC[3][3] = 0;
          break;
        case 2:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 2; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 2; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 2; pieceC[1][3] = 2; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 2; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 2; pieceC[1][2] = 2; pieceC[2][2] = 2; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 2; pieceC[2][1] = 2; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 2; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 2; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
      }
      break;
    case 3: //L
      switch (pieceR){
        case 1:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 3; pieceC[1][2] = 3; pieceC[2][2] = 3; pieceC[3][2] = 0;
          pieceC[0][3] = 3; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 2:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 3; pieceC[1][1] = 3; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 3; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 3; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 3; pieceC[3][1] = 0;
          pieceC[0][2] = 3; pieceC[1][2] = 3; pieceC[2][2] = 3; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 3; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 3; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 3; pieceC[2][3] = 3; pieceC[3][3] = 0;
          break;
      }
      break;
    case 4: //s shape
      switch (pieceR){
        case 1:
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 4; pieceC[2][2] = 4; pieceC[3][2] = 0;
          pieceC[0][3] = 4; pieceC[1][3] = 4; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 2:
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 4; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 4; pieceC[2][2] = 4; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 4; pieceC[3][3] = 0;
          break;
      }
      break;
    case 5: //T shape
      switch (pieceR){
        case 1:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 5; pieceC[1][2] = 5; pieceC[2][2] = 5; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 5; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 2:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 5; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 5; pieceC[1][2] = 5; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 5; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 5; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 5; pieceC[1][2] = 5; pieceC[2][2] = 5; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 5; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 5; pieceC[2][2] = 5; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 5; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
      }
      break;
    case 6: //reverse s
      switch (pieceR){
        case 1:
        case 3:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 0; pieceC[3][1] = 0;
          pieceC[0][2] = 6; pieceC[1][2] = 6; pieceC[2][2] = 0; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 6; pieceC[2][3] = 6; pieceC[3][3] = 0;
          break;
        case 2:
        case 4:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 0; pieceC[2][1] = 6; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 6; pieceC[2][2] = 6; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 6; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
      }
      break;
    case 7: //square
      switch (pieceR){
        default:
          pieceC[0][0] = 0; pieceC[1][0] = 0; pieceC[2][0] = 0; pieceC[3][0] = 0;
          pieceC[0][1] = 0; pieceC[1][1] = 7; pieceC[2][1] = 7; pieceC[3][1] = 0;
          pieceC[0][2] = 0; pieceC[1][2] = 7; pieceC[2][2] = 7; pieceC[3][2] = 0;
          pieceC[0][3] = 0; pieceC[1][3] = 0; pieceC[2][3] = 0; pieceC[3][3] = 0;
          break;
      }
      break;
  }
}

void nextpiece(){ //generate the next piece and move PC back to the top
  unsigned short tempT = nextpieceT;
  unsigned short tempR = nextpieceR;
  short i = 0;
  nextpieceT = rand() % 7 + 1;
  nextpieceR = rand() % 4 + 1;
  pieceT = nextpieceT;
  pieceR = nextpieceR;
  loadpiece();
  for (i = 0; i < 4; i++){
    short j = 0;
    for (j = 0; j < 4; j++){
      if (pieceC[i][j] != 0){
        setblock((i)*clspX + npX, (j)*clspY + npY);
      } else {
        clearblock((i)*clspX + npX, (j)*clspY + npY);
      }
    }
  }

  pieceT = tempT;
  pieceR = tempR;

  pieceCY = 0; //move piece carrier back to the top of screen
  pieceCX = 3;
}

void mdraw(){
  short i = 0;
  loadpiece(); //load piece into the piece carrier
  for (i = 0; i < 6; i++){ //check piece carrier and one block offset outside it
    short j = 0;
    for (j = 0; j < 6; j++){
      if ((pieceCX - 1 + i) >= 0 && (pieceCX - 1 + i) < pfsizeX && (pieceCY - 1 + j) >= 4 && (pieceCY - 1 + j) < (pfsizeY)){ //check if this segment be drawn on screen
        if (i >= 1 && i <= 4 && j >= 1 && j <= 4){ //accessing piece carrier section here}
          if (pieceC[i - 1][j - 1] > 0){ //if piece carrier block is confirmed should be white
            cdraw(i - 1 + pieceCX, j - 5 + pieceCY, pieceC[i - 1][j - 1]);
          } else {
            cdraw(i - 1 + pieceCX, j - 5 + pieceCY, playfield[i - 1 + pieceCX][j - 1 + pieceCY]); //color it in with the existing playfield background
          }
        } else {
          cdraw(i - 1 + pieceCX, j - 5 + pieceCY, playfield[i - 1 + pieceCX][j - 1 + pieceCY]);
        }
      }
    }
  }
}

unsigned short checkCollide(){ //move the piece carrier first, then check if anything collides
  unsigned short nonvalidity = 0;
  short i = 0;
  loadpiece(); //load piece into the piece carrier
  for (i = 0; i < 4; i++){ //run through all piece carrier cells
    short j = 0;
    for (j = 0; j < 4; j++){
      if (pieceCX + i >= 0 && pieceCX + i < pfsizeX && pieceCY + j >= 0 && pieceCY + j < pfsizeY){ //check if piece carrier segment can be drawn on screen
        if (pieceC[i][j] != 0 && playfield[i + pieceCX][j + pieceCY] > 7){ //if both background and nonzero piece carrier segment collide
          nonvalidity = 1;
        }
      } else { //this segment of PC can't be drawn on the screen
        if (pieceC[i][j] != 0){ //a filled in segment would be drawn offscreen
          nonvalidity = 1;
        }
      }
    }
  }
  return nonvalidity;
}

void newgamemon(){
  unsigned short i = 0;
  if (ngame == 0){
    return;
  }
  setCursor( 14, 40 ); printf("         \r\n");
  setCursor( 3, 47  ); printf("        \r\n");
  setCursor( 5, 47 );  printf("        \r\n");
  setCursor( 7, 47 );  printf("        \r\n");
  ngame = 0;
  for (i = 0; i < pfsizeX; i++){ //clear any cells with active piece parts (will be written again with new pieceC
    unsigned short j = 0;
    for (j = 0; j < pfsizeY; j++){
      playfield[i][j] = 0;
    }
  }
  lines = 0;
  level = 0;
  score = 0;
  lslvi = 0;
  fdrop = 0;
  nextpiece();
  nextpiece();
  recopyscreen();
}

void piece2bg(){ //will copy active piece into background and check for lines and game over
  unsigned short clearline[pfsizeY];
  unsigned short clearedlines = 0;
  unsigned short templines = 0;
  unsigned short i = 0;
  unsigned short j = 0;
  for (i = 0; i < 4; i++){ //copy piece onto background
    unsigned short j = 0;
    for (j = 0; j < 4; j++){
      if (pieceC[i][j] != 0){
        playfield[i + pieceCX][j + pieceCY] = pieceC[i][j] + 8; //copy the piece into the playfield/background
      }
    }
  }
  nextpiece();

  //check for line clears
  for (j = 4; j < pfsizeY; j++){
    clearline[j] = 1; //assume line is cleared
    for (i = 0; i < pfsizeX; i++){
      if (playfield[i][j] <= 7){
        clearline[j] = 0;  //line is not full
        break;
      }
    }
    clearedlines += clearline[j];
    templines += clearline[j];
  }
  if (clearedlines > 0){ //breifly animate the cleared lines, then clear them
    unsigned short f = 0;
    for (f = 0; f <= 4; f++){
      unsigned short j = 4;
      for (j = 4; j < pfsizeY; j++){
        if (clearline[j] == 1){
          unsigned short i = 0;
          for (i = 0; i < pfsizeX; i++){
            cdraw(i, j - 4, f % 2);
          }
        }
      }
	  delay(200);
    }
    for (j = 4; j < pfsizeY; j++){ //actually clear the lines
      if (clearline[j] == 1){
        unsigned short t = 0;
        for (t = j; t >= 4; t--){
          for (i = 0; i < pfsizeX; i++){
            playfield[i][t] = playfield[i][t - 1];
          }
        }
      }
    }
    lines += templines; //add number of lines to lines counter
    switch (templines){ //calculate score
      case 1:
        score += 40 * (level + 1);
        break;
      case 2:
        score += 100 * (level + 1);
        break;
      case 3:
        score += 300 * (level + 1);
        break;
      default:
        score += 1200 * (level + 1);
        break;
    }
    for (i = 1; i <= templines; i++){ //see if the level needs increasing
      lslvi++;
      if (lslvi >= 10){
        lslvi = 0;
        level++;
      }
      if (level >= maxlevel){
        level = maxlevel;
      }
    }
  }
  //check gameover (scan through line 3 and see if there are any non-active pieces in it)
  for (i = 0; i < pfsizeX; i++){
    if (playfield[i][3] > 7){
      //*********************GAME OVER********************
      setCursor( 14, 40 ); printf("GAME OVER\r\n");
      while (ngame == 0){
        controls();
      }
      newgamemon();
    }
  }

  recopyscreen();

}

void drawBorder(){ //draws a border around the playfield
  int i = 0;
  for (i = 0; i < pfsizeX; i++){
    setCursor(pfY - 1, pfX + i * clspX);
    writeCharacter(205);
    setCursor(pfY - 1, pfX + i * clspX + 1);
    writeCharacter(205);
    setCursor(pfY + pfsizeY * clspY - 4, pfX + i * clspX);
    writeCharacter(205);
    setCursor(pfY + pfsizeY * clspY - 4, pfX + i * clspX + 1);
    writeCharacter(205);
  }
  for (i = 4; i < pfsizeY; i++){
    setCursor(pfY+i-4, pfX-(1*clspX)+1);
    writeCharacter(179);
	setCursor(pfY+i-4, pfX+(pfsizeX*clspX));
    writeCharacter(179);
  }
  setCursor(pfY-1, pfX-(1*clspX)+1);
  writeCharacter(213);
  setCursor(pfY-1, pfX+(pfsizeX*clspX));
  writeCharacter(184);
  setCursor(pfY + pfsizeY * clspY - 4, pfX-(1*clspX)+1);
  writeCharacter(212);
  setCursor(pfY + pfsizeY * clspY - 4, pfX+(pfsizeX*clspX));
  writeCharacter(190);
}

void waitTick(){
	unsigned int h = 0;
	regs.h.ah = 0x00;
	int86(0x1a, &regs, &regs);
	h = regs.w.dx;
	while(h==regs.w.dx){
		controls();
		regs.h.ah = 0x00;
		int86(0x1a, &regs, &regs);
	}
}

void main(){
  clearScreen();
  setCursor(0, 40);
  printf("Wilson Tetris 2020 - IBM-PC Edition\r\n");
  setCursor( 3, 40 ); printf("Score:\r\n");
  setCursor( 5, 40 ); printf("Level:\r\n");
  setCursor( 7, 40 ); printf("Lines:\r\n");
  setCursor( 4, 55 ); printf("Next\r\n");

  setCursor( 10, 42 ); printf("J K L to move\r\n");
  setCursor( 11, 42 ); printf(" A S  to rotate\r\n");
  setCursor( 12, 42 ); printf("  @   to start new game\r\n");

  //drawb(pfX-2, pfY-1, (pfsizeX+1)*clspX, ((pfsizeY-3)*clspY)); //draw outline
  drawBorder();

  ngame = 1;
  newgamemon();
  while (1){ //*********************MAIN*PROGRAM*LOOP**************************
    unsigned short droppiece = 0;
	
	waitTick(); //this delay sets the "framerate" and also reads the controls

    dropframe++;
    if (dropframe >= (maxlevel - level)){
      dropframe = 0;
      droppiece = 1;
    }
    newgamemon();

    if (lvladd != 0){
      level += lvladd;
      lvladd = 0;
      recopyscreen();
    }

    if (newpc != 0){
      newpc = 0;
      nextpiece();
      recopyscreen();
    }

    if (rcontrol == 1){
      pieceR++; if (pieceR >= 5){
        pieceR = 1; //try to rotate piece
      }
      if (checkCollide()){
        pieceR--; if (pieceR == 0){
          pieceR = 4; //undo rotation
        }
      }
    }
    if (rcontrol == 2){
      pieceR--; if (pieceR == 0){
        pieceR = 4;
      }
      if (checkCollide()){
        pieceR++; if (pieceR >= 5){
          pieceR = 1;
        }
      }
    }
    if (dcontrol == 4){
      pieceCX--; //try and see what happens if we move the piece left
      if (checkCollide()){ //piece move is not valid
        pieceCX++; //take piece back
      }
    }
    if (dcontrol == 6){
      pieceCX++; //try and see what happens if we move the piece left
      if (checkCollide()){ //piece move is not valid
        pieceCX--; //take piece back
      }
    }
    if (dcontrol == 8){
      pieceCY--; //try and see what happens if we move the piece up
      if (checkCollide()){ //piece move is not valid
        pieceCY++; //take piece back
      }
    }
    if (dcontrol == 2 || droppiece == 1){
      pieceCY++; //try and see what happens if we move the piece down
      if (!(drepeatframe <= drepeatframes)){ //is in fast mode
        fdrop++;
      } else {
        fdrop = 0;
      }
      if (checkCollide()){ //piece move is not valid
        pieceCY--; //take piece back
        score += fdrop; //add # of fast dropped blocks to score
        fdrop = 0;
        piece2bg(); //copy piece to background and reset to next piece
      }
    }

    clearControls();
    mdraw();
	setCursor(0, 0);
  }
}
